@@ -1,7 +1,5 @@
 '''Color your output to terminal'''
-import os
 import re
-import sre_parse
 import sys
 import time
 
@@ -51,9 +49,6 @@ COLOR_TYPES = {
 # The format of a palette color
 PALETTE_COLOR_RE = re.compile(r'\b([bf])\.([a-z0-9-_]+)\b')
 
-# Detect rgb support
-RGB_SUPPORTED = os.getenv('COLORTERM') in ('truecolor', '24bit')
-
 # Select Graphic Rendition sequence (any type)
 SGR_RE = re.compile(br'\x1b\[[0-9;]*m')
 
@@ -128,7 +123,7 @@ class Color:
             # Break down hex color to RGB integers
             rgb_int = [int(hex_code[i:i + 2], 16) for i in (0, 2, 4)]
 
-            if self.rgb or (self.rgb is None and RGB_SUPPORTED):
+            if self.rgb:
                 target += b'2;'
                 color_id = b'%d;%d;%d' % tuple(rgb_int)
             else:
@@ -145,7 +140,7 @@ class Color:
             color_reset = COLOR_TYPES[style]['reset'] + color_reset
             color_types.append((style, COLOR_TYPES[style]['code']))
 
-        self._color = ' '.join(dict.fromkeys(color.lower().split()))
+        self._color = ' '.join(dict.fromkeys(color.split()))
         self.color_code = color_code
         self.color_reset = color_reset
         self.color_types = color_types
@@ -206,14 +201,14 @@ class Color:
                 else:
                     return [[source_color_code, False, None]]
 
-                colors.append([make_sgr(code), False or is_reset, color_type])
+                colors.append([make_sgr(code), is_reset, color_type])
             # Single-code SGR
             else:
                 color = [make_sgr(b'%d' % int(code)), False, None]
 
                 for name, color_type in COLOR_TYPES.items():
                     if color_type['re'].search(color[0]):
-                        color[1] = color[0] == color_type['reset'] or is_reset
+                        color[1] = is_reset or color[0] == color_type['reset']
                         color[2] = name
 
                         # Types don't overlap; only one can match
@@ -314,10 +309,10 @@ class Palette:
 
         if not re.fullmatch(r'[a-z0-9-_]+', name):
             raise ValueError('name accepts alphanumerics, dashes, and '
-                             'underscores only.')
+                             'underscores only')
 
         if not re.fullmatch(r'#[0-9a-f]{6}', value):
-            raise ValueError('color value must be in the format of `#123abc`')
+            raise ValueError('palette color must be in `#123abc` format')
 
         self.colors[name] = value
 
@@ -353,7 +348,13 @@ class Palette:
 class Rule:
     '''A rule containing a regex and colors corresponding to the regex's groups.'''
 
-    def __init__(self, regex, color=None, description=None, exclusive=False):
+    # pylint: disable=import-outside-toplevel,too-many-arguments
+    def __init__(self,
+                 regex,
+                 color=None,
+                 description=None,
+                 exclusive=False,
+                 pcre=False):
         '''Constructor.
 
         Args:
@@ -361,7 +362,8 @@ class Rule:
             color (chromaterm.Color, dict): Color used to highlight the entire
                 match. Can be a dictionary of {group:  color} format.
             description (str): String to help identify the rule.
-            exclusive (bool): Whether other rules should overlap with this one
+            exclusive (bool): Whether other rules should overlap with this one.
+            pcre (bool): Whether to use PCRE2 or default to Python's RE.
 
         Raises:
             TypeError: If `color` is not an instance of `Color` or not `None`.
@@ -379,12 +381,11 @@ class Rule:
         self.colors = {}
         self.description = description
         self.exclusive = exclusive
+        self.pcre = pcre  # Compiles regex into self._regex
 
         if not isinstance(color, dict):
             color = {0: color}
 
-        self._regex = self.compile_regex(regex.encode(encoding='utf-8'), color)
-
         for group, value in color.items():
             self.set_color(value, group)
 
@@ -397,42 +398,23 @@ class Rule:
     def color(self, value):
         self.set_color(value)
 
-    @staticmethod
-    def compile_regex(pattern, referenced_groups):
-        '''Returns an instance of `re.Pattern` after performing any optimizations
+    @property
+    def pcre(self):
+        '''True when the PCRE engine is used. False means Python's RE.'''
+        return self._pcre
 
-        Args:
-            pattern (bytes): The pattern to compile.
-            referenced_groups (list): The regex group numbers referenced.
-        '''
-        # pylint: disable=protected-access
-        original_parse_sub = sre_parse._parse_sub
-        groups = []
-        regex = None
+    @pcre.setter
+    def pcre(self, value):
+        if not isinstance(value, bool):
+            raise TypeError('pcre must be a boolean')
 
-        try:
-            if list(referenced_groups) != [0]:
-                raise re.error('Cannot optimize; groups are referenced')
+        self._pcre = value
 
-            def _parse_sub(source, *args, **kwargs):
-                # Note the start of a subpattern (i.e. group). Add in reverse
-                # to avoid calculating an index offset
-                groups.insert(0, source.tell())
-                return original_parse_sub(source, *args, **kwargs)
-
-            sre_parse._parse_sub = _parse_sub
-            regex = re.compile(pattern)
-
-            for group in groups:
-                # Group is not 0 and not a special group
-                if group and pattern[group - 1] == 40:  # b'('
-                    pattern = pattern[:group - 1] + b'(?:' + pattern[group:]
-
-            return regex if len(groups) < 2 else re.compile(pattern)
-        except (AttributeError, TypeError, re.error):
-            return regex or re.compile(pattern)
-        finally:
-            sre_parse._parse_sub = original_parse_sub
+        if value:
+            import chromaterm.pcre
+            self._regex = chromaterm.pcre.Pattern(self.regex.encode())
+        else:
+            self._regex = re.compile(self.regex.encode())
 
     def get_matches(self, data):
         '''Returns a list of tuples, each containing a start index, an end index,
@@ -473,7 +455,7 @@ class Rule:
             raise TypeError('group must be an integer or a string')
 
         if isinstance(group, str):
-            if group not in self._regex.groupindex:
+            if not self._regex.groupindex.get(group):
                 raise ValueError(f'named group {repr(group)} not in regex')
 
             # Resolve the named group to its index
@@ -487,7 +469,7 @@ class Rule:
             raise TypeError('color must be a chromaterm.Color')
 
         if group > self._regex.groups:
-            raise ValueError(f'regex only has {self._regex.groups} group(s); '
+            raise ValueError(f'regex has {self._regex.groups} group(s); '
                              f'{group} is invalid')
 
         self.colors[group] = color
@@ -568,26 +550,25 @@ class Config:
         # a test-driven approach by looking at the test_config_highlight_* tests
         for start, end, color in self.get_matches(data):
             start_index, end_index = self.get_insert_index(start, end, inserts)
-            reset_codes = []
 
             # Each color type requires tracking of its respective type
             for color_type, color_code in color.color_types:
-                start_insert = [start, color_code, False, color_type]
-                end_insert = [
-                    end, self._reset_codes[color_type], True, color_type
-                ]
-
                 # Find the last color before the end of this match (if any) and
                 # use it as the reset code for this color
                 for insert in inserts[end_index:]:
                     if insert[3] == color_type:
-                        end_insert[1] = insert[1]
+                        reset = insert[1]
                         break
 
                     # No type (a full reset); use the default for this type
                     if insert[2] and insert[3] is None:
-                        end_insert[1] = COLOR_TYPES[color_type]['reset']
+                        reset = COLOR_TYPES[color_type]['reset']
                         break
+                else:
+                    reset = self._reset_codes[color_type]
+
+                start_insert = [start, color_code, False, color_type]
+                end_insert = [end, reset, True, color_type]
 
                 # Replace every color reset of the current color type with our
                 # color code to prevent them from interrupting this color
@@ -599,13 +580,14 @@ class Config:
 
                         insert[1:4] = color_code, False, color_type
 
-                # After all of the start inserts are added, the end ones can
-                # placed in reverse order (outward from the match).
+                # Relative to data, the inserts are added in reverse order LI-FO
                 inserts.insert(start_index, start_insert)
-                reset_codes.insert(0, end_insert)
+                inserts.insert(end_index, end_insert)
 
-            for reset_code in reset_codes:
-                inserts.insert(end_index, reset_code)
+                # Advance to ensure the slices above search appropriately if
+                # multiple color types exist
+                start_index += 1
+                end_index += 1
 
         return inserts
 
