@@ -7,6 +7,7 @@ import re
 import select
 import signal
 import sys
+from ctypes.util import find_library
 
 import yaml
 
@@ -27,7 +28,7 @@ CONFIG_LOCATIONS = [
 CWD_UPDATE_INTERVAL = 1 / 4
 
 # Maximum chuck size per read
-READ_SIZE = 4096  # 4 KiB
+READ_SIZE = 8192
 
 # Sequences upon which ct will split during processing (ECMA 035 and 048):
 # * new lines, vertical spaces, form feeds;
@@ -42,9 +43,10 @@ SPLIT_RE = re.compile(
     br'\x1b\x5b[\x30-\x3f]*[\x20-\x2f]*[\x40-\x6c\x6e-\x7e]|'
     br'\x1b[\x50\x58\x5d\x5e\x5f][^\x07\x1b]*(?:\x07|\x1b\x5c)?)')
 
-# The start of a control string
-SPLIT_CONTROL_STRINGS = (b'\x1b\x50', b'\x1b\x58', b'\x1b\x5d', b'\x1b\x5e',
-                         b'\x1b\x5f')
+# Control strings that have arbitrary lengths
+ANSI_CONTROL_STRINGS_START = (b'\x1b\x50', b'\x1b\x58', b'\x1b\x5d',
+                              b'\x1b\x5e', b'\x1b\x5f')
+ANSI_CONTROL_STRINGS_END = (b'\x07', b'\x1b\x5c')
 
 
 def args_init(args=None):
@@ -53,7 +55,8 @@ def args_init(args=None):
     Args:
         args (list): A list of program arguments, Defaults to sys.argv.
     '''
-    parser = argparse.ArgumentParser()
+    formatter = lambda prog: argparse.HelpFormatter(prog, max_help_position=30)
+    parser = argparse.ArgumentParser(formatter_class=formatter)
     parser.epilog = 'For more info, go to https://github.com/hSaria/ChromaTerm.'
 
     parser.add_argument('program',
@@ -66,25 +69,39 @@ def args_init(args=None):
                         nargs=argparse.REMAINDER,
                         help=argparse.SUPPRESS)
 
-    parser.add_argument('--benchmark',
+    parser.add_argument('-b',
+                        '--benchmark',
                         action='store_true',
                         help='at exit, print rule usage statistics')
 
-    parser.add_argument('--config',
-                        metavar='FILE',
+    parser.add_argument('-c',
+                        '--config',
+                        metavar='file',
                         help='override config file location (default: ~/'
                         '.chromaterm.yml)')
 
-    parser.add_argument('--reload',
+    parser.add_argument('-r',
+                        '--reload',
                         action='store_true',
                         help='reload the config of all CT instances')
 
-    parser.add_argument('--rgb',
+    parser.add_argument('-R',
+                        '--rgb',
                         action='store_true',
                         help='use RGB colors (default: detect support, '
                         'fallback to xterm-256)')
 
-    return parser.parse_args(args=args)
+    parser.add_argument('--pcre',
+                        action='store_true',
+                        help="(advanced) use PCRE2 instead of Python's RE"
+                        if find_library('pcre2-8') else argparse.SUPPRESS)
+
+    args = parser.parse_args(args=args)
+
+    # Detect rgb support
+    args.rgb = args.rgb or os.getenv('COLORTERM') in ('truecolor', '24bit')
+
+    return args
 
 
 def eprint(*args, **kwargs):
@@ -95,8 +112,7 @@ def eprint(*args, **kwargs):
 
 def get_default_config_location():
     '''Returns the first location in `CONFIG_LOCATIONS` that points to a file,
-    defaulting to the first item in the list.
-    '''
+    defaulting to the first item in the list.'''
     resolve = lambda x: os.path.expanduser(os.path.expandvars(x))
 
     for location in CONFIG_LOCATIONS:
@@ -133,7 +149,7 @@ def get_wait_duration(buffer, min_wait=1 / 256, max_wait=1 / 8):
     return min_wait
 
 
-def load_config(config, data, rgb=False):
+def load_config(config, data, rgb=False, pcre=False):
     '''Reads configuration from a YAML-based string, formatted like so:
     palette:
       red: "#ff0000"
@@ -187,7 +203,7 @@ def load_config(config, data, rgb=False):
     config.rules.clear()
 
     for rule in rules:
-        rule = parse_rule(rule, palette=palette, rgb=rgb)
+        rule = parse_rule(rule, palette=palette, rgb=rgb, pcre=pcre)
 
         if isinstance(rule, Rule):
             config.rules.append(rule)
@@ -217,7 +233,7 @@ def parse_palette(data):
     return palette
 
 
-def parse_rule(data, palette=None, rgb=False):
+def parse_rule(data, palette=None, rgb=False, pcre=False):
     '''Returns an instance of chromaterm.Rule if parsed correctly. Otherwise, a
     string with the error message is returned.
 
@@ -246,7 +262,7 @@ def parse_rule(data, palette=None, rgb=False):
         for group, value in color.items():
             color[group] = Color(value, palette=palette, rgb=rgb)
 
-        return Rule(regex, color, description, exclusive)
+        return Rule(regex, color, description, exclusive, pcre)
     except (TypeError, ValueError) as exception:
         return f'Error on {rule_repr}: {exception}'
     except re.error as exception:
@@ -319,16 +335,18 @@ def process_input(config, data_fd, forward_fd=None, max_wait=None):
 
             data, separator = chunks[-1]
 
-            # Separator is an incomplete control strings; wait for the rest
-            if data_read and separator.startswith(SPLIT_CONTROL_STRINGS):
+            # Separator is an incomplete control string; wait for the rest
+            if data_read and separator.startswith(ANSI_CONTROL_STRINGS_START) \
+                    and not separator.endswith(ANSI_CONTROL_STRINGS_END):
                 buffer = data + separator
-            # Zero or one characters indicates keyboard typing; don't highlight
-            # Account for the backspaces added by some shells, like zsh
-            elif len(data) < 2 + data.count(b'\b') * 2:
+            # A single character indicates keyboard typing; don't highlight
+            # Account for backspaces added by some shells, like zsh
+            elif 0 < len(data_read) < 2 + data_read.count(b'\b') * 2:
                 sys.stdout.buffer.write(data + separator)
                 buffer = b''
-            # There's more data; fetch it before highlighting
-            elif data_read and read_ready(*fds, timeout=wait_duration):
+            # There's more data; fetch it before highlighting, if buffer isn't full
+            elif data_read and len(buffer) <= READ_SIZE \
+                    and read_ready(*fds, timeout=wait_duration):
                 buffer = data + separator
             else:
                 sys.stdout.buffer.write(config.highlight(data) + separator)
@@ -390,7 +408,6 @@ def reload_chromaterm_instances():
             if process.cmdline()[:2] == current_process.cmdline()[:2]:
                 os.kill(process.pid, signal.SIGUSR1)
                 count += 1
-        # As per the docs, expect those errors when accessing the process methods
         except (psutil.AccessDenied, psutil.NoSuchProcess):
             pass
 
@@ -419,7 +436,7 @@ def run_program(program_args):
         attributes = termios.tcgetattr(sys.stdin.fileno())
         atexit.register(termios.tcsetattr, sys.stdin.fileno(), termios.TCSANOW,
                         attributes)
-        tty.setraw(sys.stdin.fileno())
+        tty.setraw(sys.stdin.fileno(), termios.TCSANOW)
     except termios.error:
         attributes = None
 
@@ -433,8 +450,8 @@ def run_program(program_args):
 
         try:
             os.execvp(program_args[0], program_args)
-        except FileNotFoundError:
-            eprint(f'command not found: {program_args[0]}')
+        except OSError as exception:
+            eprint(f'{exception.strerror.lower()}: {program_args[0]}')
 
         # exec replaces the fork's process; only hit on exception
         sys.exit(1)
@@ -512,8 +529,8 @@ def main(args=None, max_wait=None, write_default=True):
 
         # Write default config if not there
         if write_default and not os.access(args.config, os.F_OK):
-            from chromaterm.default_config import write_default_config
-            write_default_config(args.config)
+            import chromaterm.default_config
+            chromaterm.default_config.write_default_config(args.config)
 
     config = Config(benchmark=args.benchmark)
 
@@ -521,16 +538,6 @@ def main(args=None, max_wait=None, write_default=True):
     if args.benchmark:
         atexit.register(config.print_benchmark_results)
 
-    # Create the signal handler to trigger reloading the config
-    def reload_config_handler(*_):
-        config_data = read_file(args.config)
-
-        if config_data:
-            load_config(config, config_data, rgb=args.rgb or None)
-
-    # Trigger the initial loading
-    reload_config_handler()
-
     if args.program:
         # ChromaTerm is spawning the program in a controlling terminal; stdin is
         # being forwarded to the program
@@ -541,6 +548,16 @@ def main(args=None, max_wait=None, write_default=True):
         data_fd = sys.stdin.fileno()
         forward_fd = None
 
+    # Create the signal handler to trigger reloading the config
+    def reload_config_handler(*_):
+        config_data = read_file(args.config)
+
+        if config_data:
+            load_config(config, config_data, rgb=args.rgb, pcre=args.pcre)
+
+    # Trigger the initial loading
+    reload_config_handler()
+
     # Ignore SIGINT (CTRL+C) and attach reload handler
     signal.signal(signal.SIGINT, signal.SIG_IGN)
     signal.signal(signal.SIGUSR1, reload_config_handler)
@@ -552,7 +569,7 @@ def main(args=None, max_wait=None, write_default=True):
         # Surpress the implicit flush that Python runs on exit
         sys.stdout.flush = lambda: None
 
-    return None
+    return os.wait()[1] >> 8 if args.program else 0
 
 
 if __name__ == '__main__':
